/** @file function.cpp
 *
 *  Implementation of class of symbolic functions. */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <string>
#include <vector>

#include "function.h"

#include "inifcns.h"
#include "operators.h"
#include "power.h"
#include "utils.h"

namespace GiNaC {

GINAC_IMPLEMENT_REGISTERED_CLASS(function, exprseq)

function::function(const archive_node& n, lst& sym_lst) : inherited(n, sym_lst)
{
}

void function::archive(archive_node& n) const
{ 
    inherited::archive(n);
}

ex function::unarchive(const archive_node& n, lst& sym_lst)
{
    return (new function(n, sym_lst))->setflag(status_flags::dynallocated);
}

function* function::duplicate() const
{
	return new function(*this);
}

void function::accept(GiNaC::visitor& v) const
{
	if (visitor* p = dynamic_cast<visitor*>(&v)) {
		p->visit(*this);
	} else {
		inherited::accept(v);
	}
}

int function::compare_same_type(const basic& other) const 
{
	if (tinfo_key == other.tinfo()) {
		return exprseq::compare_same_type(other); 
	} else {
		return compare_pointers(tinfo_key, other.tinfo());
	}
}

bool function::is_equal_same_type(const basic& other) const 
{
	if (tinfo_key == other.tinfo()) {
		return exprseq::is_equal_same_type(other);
	} else {
		return false;
	}
}

bool function::match_same_type(const basic & other) const
{
	return tinfo_key == other.tinfo();
}

unsigned function::return_type() const
{
//	GINAC_ASSERT(serial<registered_functions().size());
	// Default behavior is to use the return type of the first
	// argument. Thus, exp() of a matrix behaves like a matrix, etc.
	if (seq.empty()) {
		return return_types::commutative;
	} else {
		return seq.begin()->return_type();
	}
}

tinfo_t function::return_type_tinfo() const
{
	//GINAC_ASSERT(serial<registered_functions().size());
	// Default behavior is to use the return type of the first
	// argument. Thus, exp() of a matrix behaves like a matrix, etc.
	if (seq.empty()) {
		return this;
	} else {
		return seq.begin()->return_type_tinfo();
	}
}

ex function::conjugate() const
{
	return conjugate_function(*this);
}

ex function::derivative(const symbol& s) const
{
	// Chain rule
	ex arg_diff;
	ex result;
	size_t num = seq.size();
	for (size_t i=0; i<num; i++) {
		arg_diff = seq[i].diff(s);
		// We apply the chain rule only when it makes sense.  This is not
		// just for performance reasons but also to allow functions to
		// throw when differentiated with respect to one of its arguments
		// without running into trouble with our automatic full
		// differentiation:
		if (!arg_diff.is_zero()) {
			result += pderivative(i) * arg_diff;
		}
	}
	return result;
}

ex function::pderivative(unsigned diff_param) const
{
	return function_derivative(lst(diff_param), *this).hold();
}

ex function::power_law(const ex& exp) const
{
	return power::power(*this, exp).hold();
}

ex function::expand(unsigned options) const
{
	return setflag(status_flags::expanded);
//TODO??
//    // Only expand arguments when asked to do so
//    if (options & expand_options::expand_function_args)
//        return inherited::expand(options);
//    else
//        return (options == 0) ? setflag(status_flags::expanded) : *this;
}

bool function::info(unsigned inf) const
{
	if (inf == info_flags::function) {
		return true;
	} else {
		return inherited::info(inf);
	}
}

void function::print(const print_context& c, unsigned level) const
{
	const std::vector<print_functor>& pdt = get_class_info().options.get_print_dispatch_table();
	unsigned id = c.get_class_info().options.get_id();
	if (id >= pdt.size() || !(pdt[id].is_valid())) {
		if (is_a<print_tree>(c)) {
			c.s << std::string(level, ' ') << class_name()
				<< " @" << this
				<< std::hex << ", hash=0x" << hashvalue << ", flags=0x" << flags << std::dec
				<< ", nops=" << nops()
				<< std::endl;
			unsigned delta_indent = static_cast<const print_tree&>(c).delta_indent;
			for (size_t i=0; i<seq.size(); ++i) {
				seq[i].print(c, level + delta_indent);
			}
			c.s << std::string(level + delta_indent, ' ') << "=====" << std::endl;
		} else if (is_a<print_latex>(c)) {
			c.s << "\\mbox{" << class_name() << "}";
			inherited::do_print(c,level);
		} else {
			std::string classname(class_name());
			c.s << classname.erase(classname.find("_function",0),9);
			inherited::do_print(c,level);
		}
	} else {
		pdt[id](*this, c, level);
	}
}

} // namespace GiNaC
