/** @file inifcns_exp.cpp
 *
 *  Implementation of TODO
 *  functions. */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "inifcns_exp.h"

#include <vector>
#include <stdexcept>

#include "inifcns.h"
#include "inifcns_polylog.h"
#include "ex.h"
#include "constant.h"
#include "numeric.h"
#include "power.h"
#include "operators.h"
#include "relational.h"
#include "symbol.h"
#include "symmetry.h"
#include "pseries.h"
#include "utils.h"

namespace GiNaC {

GINAC_IMPLEMENT_FUNCTION_OPT(eta_function,
		print_func<print_latex>(&eta_function::do_print_latex))

ex eta_function::conjugate() const
{
	return -eta(seq[0], seq[1]);
}

ex eta_function::eval(int level) const
{
	// Canonicalize argument order according to the symmetry properties
	exvector v = seq;
	int sig = canonicalize(v.begin(), sy_symm(0,1));
	if (sig != INT_MAX) {
		// Something has changed while sorting arguments, more evaluations later
		if (sig == 0) {
			return _ex0;
		}
		return ex(sig) * thiscontainer(v);
	}

	const ex& x = seq[0];
	const ex& y = seq[1];
	// trivial:  eta(x,c) -> 0  if c is real and positive
	if (x.info(info_flags::positive) || y.info(info_flags::positive))
		return _ex0;

	if (x.info(info_flags::numeric) &&	y.info(info_flags::numeric)) {
		// don't call eta_evalf here because it would call Pi.evalf()!
		const numeric nx = ex_to<numeric>(x);
		const numeric ny = ex_to<numeric>(y);
		const numeric nxy = ex_to<numeric>(x*y);
		int cut = 0;
		if (nx.is_real() && nx.is_negative())
			cut -= 4;
		if (ny.is_real() && ny.is_negative())
			cut -= 4;
		if (nxy.is_real() && nxy.is_negative())
			cut += 4;
		return (I/4)*Pi*((csgn(-imag(nx))+1)*(csgn(-imag(ny))+1)*(csgn(imag(nxy))+1)-
		                 (csgn(imag(nx))+1)*(csgn(imag(ny))+1)*(csgn(-imag(nxy))+1)+cut);
	}
	
	return this->hold();
}

ex eta_function::evalf(int level) const
{
	const ex& x = seq[0];
	const ex& y = seq[1];
	// It seems like we basically have to replicate the eval function here,
	// since the expression might not be fully evaluated yet.
	if (x.info(info_flags::positive) || y.info(info_flags::positive))
		return _ex0;

	if (x.info(info_flags::numeric) &&	y.info(info_flags::numeric)) {
		const numeric nx = ex_to<numeric>(x);
		const numeric ny = ex_to<numeric>(y);
		const numeric nxy = ex_to<numeric>(x*y);
		int cut = 0;
		if (nx.is_real() && nx.is_negative())
			cut -= 4;
		if (ny.is_real() && ny.is_negative())
			cut -= 4;
		if (nxy.is_real() && nxy.is_negative())
			cut += 4;
		return GiNaC::evalf(I/4*Pi)*((csgn(-imag(nx))+1)*(csgn(-imag(ny))+1)*(csgn(imag(nxy))+1)-
		                      (csgn(imag(nx))+1)*(csgn(imag(ny))+1)*(csgn(-imag(nxy))+1)+cut);
	}

	return this->hold();
}

ex eta_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& x = seq[0];
	const ex& y = seq[1];
	const ex x_pt = x.subs(rel, subs_options::no_pattern);
	const ex y_pt = y.subs(rel, subs_options::no_pattern);
	if ((x_pt.info(info_flags::numeric) && x_pt.info(info_flags::negative)) ||
	    (y_pt.info(info_flags::numeric) && y_pt.info(info_flags::negative)) ||
	    ((x_pt*y_pt).info(info_flags::numeric) && (x_pt*y_pt).info(info_flags::negative)))
			throw (std::domain_error("eta_series(): on discontinuity"));
	epvector seq;
	seq.push_back(expair(eta(x_pt,y_pt), _ex0));
	return pseries(rel, seq);
}

void eta_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\eta";
	inherited::do_print(c,level);
}

//////////
// exponential function
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(exp_function,
		print_func<print_csrc_float>(&exp_function::do_print_csrc_float).
		print_func<print_latex>(&exp_function::do_print_latex))

ex exp_function::eval(int level) const
{
	const ex& x = seq[0];

	// exp(0) -> 1
	if (x.is_zero()) {
		return _ex1;
	}

	// exp(n*Pi*I/2) -> {+1|+I|-1|-I}
	const ex TwoExOverPiI=(_ex2*x)/(Pi*I);
	if (TwoExOverPiI.info(info_flags::integer)) {
		const numeric z = mod(ex_to<numeric>(TwoExOverPiI),*_num4_p);
		if (z.is_equal(*_num0_p))
			return _ex1;
		if (z.is_equal(*_num1_p))
			return ex(I);
		if (z.is_equal(*_num2_p))
			return _ex_1;
		if (z.is_equal(*_num3_p))
			return ex(-I);
	}

	// exp(log(x)) -> x
	if (is_exactly_a<log_function>(x))
		return x.op(0);
	
	// exp(float) -> float
	if (x.info(info_flags::numeric) && !x.info(info_flags::crational))
		return exp(ex_to<numeric>(x));
	
	return this->hold();
}

ex exp_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return exp(ex_to<numeric>(x));
	
	return this->hold();
}

ex exp_function::pderivative(unsigned deriv_param) const
{
	GINAC_ASSERT(deriv_param==0);

	// d/dx exp(x) -> exp(x)
	return *this;
}

void exp_function::do_print_csrc_float(const print_context& c, unsigned level) const
{
	c.s << "exp("; seq[0].print(c); c.s << ")";
}

void exp_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\exp";
	inherited::do_print(c,level);
}

//////////
// natural logarithm
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(log_function,
		print_func<print_csrc_float>(&log_function::do_print_csrc_float).
		print_func<print_latex>(&log_function::do_print_latex))

ex log_function::eval(int level) const
{
	const ex& x = seq[0];
	if (x.info(info_flags::numeric)) {
		if (x.is_zero())         // log(0) -> infinity
			throw(pole_error("log_eval(): log(0)",0));
		if (x.info(info_flags::rational) && x.info(info_flags::negative))
			return (log(-x)+I*Pi);
		if (x.is_equal(_ex1))  // log(1) -> 0
			return _ex0;
		if (x.is_equal(I))       // log(I) -> Pi*I/2
			return (Pi*I*_ex1_2);
		if (x.is_equal(-I))      // log(-I) -> -Pi*I/2
			return (Pi*I*_ex_1_2);

		// log(float) -> float
		if (!x.info(info_flags::crational))
			return log(ex_to<numeric>(x));
	}

	// log(exp(t)) -> t (if -Pi < t.imag() <= Pi):
	if (is_exactly_a<exp_function>(x)) {
		const ex& t = x.op(0);
		if (t.info(info_flags::real)) {
			return t;
		}
	}
	
	return this->hold();
}

ex log_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return log(ex_to<numeric>(x));
	
	return this->hold();
}

ex log_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx log(x) -> 1/x
	return power::power(x, _ex_1);
}

ex log_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];

	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	ex arg_pt;
	bool must_expand_arg = false;
	// maybe substitution of rel into arg fails because of a pole
	try {
		arg_pt = arg.subs(rel, subs_options::no_pattern);
	} catch (pole_error) {
		must_expand_arg = true;
	}
	// or we are at the branch point anyways
	if (arg_pt.is_zero())
		must_expand_arg = true;
	
	if (must_expand_arg) {
		// method:
		// This is the branch point: Series expand the argument first, then
		// trivially factorize it to isolate that part which has constant
		// leading coefficient in this fashion:
		//   x^n + x^(n+1) +...+ Order(x^(n+m))  ->  x^n * (1 + x +...+ Order(x^m)).
		// Return a plain n*log(x) for the x^n part and series expand the
		// other part.  Add them together and reexpand again in order to have
		// one unnested pseries object.  All this also works for negative n.
		pseries argser;          // series expansion of log's argument
		unsigned extra_ord = 0;  // extra expansion order
		do {
			// oops, the argument expanded to a pure Order(x^something)...
			argser = ex_to<pseries>(arg.series(rel, order+extra_ord, options));
			++extra_ord;
		} while (!argser.is_terminating() && argser.nops()==1);

		const symbol &s = ex_to<symbol>(rel.lhs());
		const ex &point = rel.rhs();
		const int n = argser.ldegree(s);
		epvector seq;
		// construct what we carelessly called the n*log(x) term above
		const ex coeff = argser.coeff(s, n);
		// expand the log, but only if coeff is real and > 0, since otherwise
		// it would make the branch cut run into the wrong direction
		if (coeff.info(info_flags::positive))
			seq.push_back(expair(n*log(s-point)+log(coeff), _ex0));
		else
			seq.push_back(expair(log(coeff*pow(s-point, n)), _ex0));

		if (!argser.is_terminating() || argser.nops()!=1) {
			// in this case n more (or less) terms are needed
			// (sadly, to generate them, we have to start from the beginning)
			if (n == 0 && coeff == 1) {
				epvector epv;
				ex acc = (new pseries(rel, epv))->setflag(status_flags::dynallocated);
				epv.reserve(2);
				epv.push_back(expair(-1, _ex0));
				epv.push_back(expair(Order(_ex1), order));
				ex rest = pseries(rel, epv).add_series(argser);
				for (int i = order-1; i>0; --i) {
					epvector cterm;
					cterm.reserve(1);
					cterm.push_back(expair(i%2 ? _ex1/i : _ex_1/i, _ex0));
					acc = pseries(rel, cterm).add_series(ex_to<pseries>(acc));
					acc = (ex_to<pseries>(rest)).mul_series(ex_to<pseries>(acc));
				}
				return acc;
			}
			const ex newarg = ex_to<pseries>((arg/coeff).series(rel, order+n, options)).shift_exponents(-n).convert_to_poly(true);
			return pseries(rel, seq).add_series(ex_to<pseries>(log(newarg).series(rel, order, options)));
		} else  // it was a monomial
			return pseries(rel, seq);
	}
	if (!(options & series_options::suppress_branchcut) &&
	     arg_pt.info(info_flags::negative)) {
		// method:
		// This is the branch cut: assemble the primitive series manually and
		// then add the corresponding complex step function.
		const symbol &s = ex_to<symbol>(rel.lhs());
		const ex &point = rel.rhs();
		const symbol foo;
		const ex replarg = log(arg).series(s==foo, order).subs(foo==point, subs_options::no_pattern);
		epvector seq;
		seq.push_back(expair(-I*csgn(arg*I)*Pi, _ex0));
		seq.push_back(expair(Order(_ex1), order));
		return (replarg - I*Pi + pseries(rel, seq)).series(rel, order);
	}
	return basic::series(rel, order, options);
}

void log_function::do_print_csrc_float(const print_context& c, unsigned level) const
{
	c.s << "log("; seq[0].print(c); c.s << ")";
}

void log_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\log";
	inherited::do_print(c,level);
}

//////////
// Logarithm of Gamma function
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(lgamma_function,
		print_func<print_latex>(&lgamma_function::do_print_latex))

/** Evaluation of lgamma(x), the natural logarithm of the Gamma function.
 *  Knows about integer arguments and that's it.  Somebody ought to provide
 *  some good numerical evaluation some day...
 *
 *  @exception GiNaC::pole_error("lgamma_eval(): logarithmic pole",0) */
ex lgamma_function::eval(int level) const
{
	const ex& x = seq[0];
	if (x.info(info_flags::numeric)) {
		// trap integer arguments:
		if (x.info(info_flags::integer)) {
			// lgamma(n) -> log((n-1)!) for postitive n
			if (x.info(info_flags::posint))
				return log(factorial(x + _ex_1));
			else
				throw (pole_error("lgamma_eval(): logarithmic pole",0));
		}
		//  lgamma_evalf should be called here once it becomes available
	}
	
	return this->hold();
}

ex lgamma_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x)) {
		try {
			return lgamma(ex_to<numeric>(x));
		} catch (const dunno& e) { }
	}
	
	return this->hold();
}

ex lgamma_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx  lgamma(x) -> psi(x)
	return psi(x);
}

ex lgamma_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];
	// method:
	// Taylor series where there is no pole falls back to psi function
	// evaluation.
	// On a pole at -m we could use the recurrence relation
	//   lgamma(x) == lgamma(x+1)-log(x)
	// from which follows
	//   series(lgamma(x),x==-m,order) ==
	//   series(lgamma(x+m+1)-log(x)...-log(x+m)),x==-m,order);
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!arg_pt.info(info_flags::integer) || arg_pt.info(info_flags::positive))
		return basic::series(rel, order, options);
	// if we got here we have to care for a simple pole of tgamma(-m):
	numeric m = -ex_to<numeric>(arg_pt);
	ex recur;
	for (numeric p = 0; p<=m; ++p)
		recur += log(arg+p);
	return (lgamma(arg+m+_ex1)-recur).series(rel, order, options);
}

void lgamma_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\log \\Gamma";
	inherited::do_print(c,level);
}

//////////
// true Gamma function
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(tgamma_function,
		print_func<print_latex>(&tgamma_function::do_print_latex))

/** Evaluation of tgamma(x), the true Gamma function.  Knows about integer
 *  arguments, half-integer arguments and that's it. Somebody ought to provide
 *  some good numerical evaluation some day...
 *
 *  @exception pole_error("tgamma_eval(): simple pole",0) */
ex tgamma_function::eval(int level) const
{
	const ex& x = seq[0];
	if (x.info(info_flags::numeric)) {
		// trap integer arguments:
		const numeric two_x = (*_num2_p)*ex_to<numeric>(x);
		if (two_x.is_even()) {
			// tgamma(n) -> (n-1)! for postitive n
			if (two_x.is_positive()) {
				return factorial(ex_to<numeric>(x).sub(*_num1_p));
			} else {
				throw (pole_error("tgamma_eval(): simple pole",1));
			}
		}
		// trap half integer arguments:
		if (two_x.is_integer()) {
			// trap positive x==(n+1/2)
			// tgamma(n+1/2) -> Pi^(1/2)*(1*3*..*(2*n-1))/(2^n)
			if (two_x.is_positive()) {
				const numeric n = ex_to<numeric>(x).sub(*_num1_2_p);
				return (doublefactorial(n.mul(*_num2_p).sub(*_num1_p)).div(pow(*_num2_p,n))) * sqrt(Pi);
			} else {
				// trap negative x==(-n+1/2)
				// tgamma(-n+1/2) -> Pi^(1/2)*(-2)^n/(1*3*..*(2*n-1))
				const numeric n = abs(ex_to<numeric>(x).sub(*_num1_2_p));
				return (pow(*_num_2_p, n).div(doublefactorial(n.mul(*_num2_p).sub(*_num1_p))))*sqrt(Pi);
			}
		}
		//  tgamma_evalf should be called here once it becomes available
	}
	
	return this->hold();
}

ex tgamma_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x)) {
		try {
			return tgamma(ex_to<numeric>(x));
		} catch (const dunno &e) { }
	}
	
	return this->hold();
}

ex tgamma_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx  tgamma(x) -> psi(x)*tgamma(x)
	return psi(x)*tgamma(x);
}

ex tgamma_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];
	// method:
	// Taylor series where there is no pole falls back to psi function
	// evaluation.
	// On a pole at -m use the recurrence relation
	//   tgamma(x) == tgamma(x+1) / x
	// from which follows
	//   series(tgamma(x),x==-m,order) ==
	//   series(tgamma(x+m+1)/(x*(x+1)*...*(x+m)),x==-m,order);
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!arg_pt.info(info_flags::integer) || arg_pt.info(info_flags::positive)) {
		return basic::series(rel, order, options);
	}
	// if we got here we have to care for a simple pole at -m:
	const numeric m = -ex_to<numeric>(arg_pt);
	ex ser_denom = _ex1;
	for (numeric p; p<=m; ++p)
		ser_denom *= arg+p;
	return (tgamma(arg+m+_ex1)/ser_denom).series(rel, order, options);
}

void tgamma_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\Gamma";
	inherited::do_print(c,level);
}

//////////
// beta-function
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(beta_function,
		print_func<print_latex>(&beta_function::do_print_latex))

ex beta_function::eval(int level) const
{
	// Canonicalize argument order according to the symmetry properties
	exvector v = seq;
	int sig = canonicalize(v.begin(), sy_symm(0,1));
	if (sig != INT_MAX) {
		// Something has changed while sorting arguments, more evaluations later
		if (sig == 0) {
			return _ex0;
		}
		return ex(sig) * thiscontainer(v);
	}

	const ex& x = seq[0];
	const ex& y = seq[1];
	if (x.is_equal(_ex1))
		return 1/y;
	if (y.is_equal(_ex1))
		return 1/x;
	if (x.info(info_flags::numeric) && y.info(info_flags::numeric)) {
		// treat all problematic x and y that may not be passed into tgamma,
		// because they would throw there although beta(x,y) is well-defined
		// using the formula beta(x,y) == (-1)^y * beta(1-x-y, y)
		const numeric &nx = ex_to<numeric>(x);
		const numeric &ny = ex_to<numeric>(y);
		if (nx.is_real() && nx.is_integer() &&
			ny.is_real() && ny.is_integer()) {
			if (nx.is_negative()) {
				if (nx<=-ny)
					return pow(*_num_1_p, ny)*beta(1-x-y, y);
				else
					throw (pole_error("beta_eval(): simple pole",1));
			}
			if (ny.is_negative()) {
				if (ny<=-nx)
					return pow(*_num_1_p, nx)*beta(1-y-x, x);
				else
					throw (pole_error("beta_eval(): simple pole",1));
			}
			return tgamma(x)*tgamma(y)/tgamma(x+y);
		}
		// no problem in numerator, but denominator has pole:
		if ((nx+ny).is_real() &&
		    (nx+ny).is_integer() &&
		   !(nx+ny).is_positive())
			 return _ex0;
		// beta_evalf should be called here once it becomes available
	}
	
	return this->hold();
}

ex beta_function::evalf(int level) const
{
	const ex& x = seq[0];
	const ex& y = seq[1];
	if (is_exactly_a<numeric>(x) && is_exactly_a<numeric>(y)) {
		try {
			return tgamma(ex_to<numeric>(x))*tgamma(ex_to<numeric>(y))/tgamma(ex_to<numeric>(x+y));
		} catch (const dunno &e) { }
	}
	
	return this->hold();
}

ex beta_function::pderivative(unsigned deriv_param) const

{
	const ex& x = seq[0];
	const ex& y = seq[1];
	GINAC_ASSERT(deriv_param<2);
	ex retval;
	
	// d/dx beta(x,y) -> (psi(x)-psi(x+y)) * beta(x,y)
	if (deriv_param==0)
		retval = (psi(x)-psi(x+y))*beta(x,y);
	// d/dy beta(x,y) -> (psi(y)-psi(x+y)) * beta(x,y)
	if (deriv_param==1)
		retval = (psi(y)-psi(x+y))*beta(x,y);
	return retval;
}

ex beta_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg1 = seq[0];
	const ex& arg2 = seq[1];
	// method:
	// Taylor series where there is no pole of one of the tgamma functions
	// falls back to beta function evaluation.  Otherwise, fall back to
	// tgamma series directly.
	const ex arg1_pt = arg1.subs(rel, subs_options::no_pattern);
	const ex arg2_pt = arg2.subs(rel, subs_options::no_pattern);
	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	const symbol &s = ex_to<symbol>(rel.lhs());
	ex arg1_ser, arg2_ser, arg1arg2_ser;
	if ((!arg1_pt.info(info_flags::integer) || arg1_pt.info(info_flags::positive)) &&
	    (!arg2_pt.info(info_flags::integer) || arg2_pt.info(info_flags::positive)))
		return basic::series(rel, order, options);
	// trap the case where arg1 is on a pole:
	if (arg1.info(info_flags::integer) && !arg1.info(info_flags::positive))
		arg1_ser = tgamma(arg1+s);
	else
		arg1_ser = tgamma(arg1);
	// trap the case where arg2 is on a pole:
	if (arg2.info(info_flags::integer) && !arg2.info(info_flags::positive))
		arg2_ser = tgamma(arg2+s);
	else
		arg2_ser = tgamma(arg2);
	// trap the case where arg1+arg2 is on a pole:
	if ((arg1+arg2).info(info_flags::integer) && !(arg1+arg2).info(info_flags::positive))
		arg1arg2_ser = tgamma(arg2+arg1+s);
	else
		arg1arg2_ser = tgamma(arg2+arg1);
	// compose the result (expanding all the terms):
	return (arg1_ser*arg2_ser/arg1arg2_ser).series(rel, order, options).expand();
}

void beta_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\mbox{B}";
	inherited::do_print(c,level);
}

//////////
// Psi-function (aka digamma-function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(psi1_function,
		print_func<print_latex>(&psi1_function::do_print_latex))

/** Evaluation of digamma-function psi(x).
 *  Somebody ought to provide some good numerical evaluation some day... */
ex psi1_function::eval(int level) const
{
	const ex& x = seq[0];
	if (x.info(info_flags::numeric)) {
		const numeric &nx = ex_to<numeric>(x);
		if (nx.is_integer()) {
			// integer case 
			if (nx.is_positive()) {
				// psi(n) -> 1 + 1/2 +...+ 1/(n-1) - Euler
				numeric rat = 0;
				for (numeric i(nx+(*_num_1_p)); i>0; --i)
					rat += i.inverse();
				return rat-Euler;
			} else {
				// for non-positive integers there is a pole:
				throw (pole_error("psi_eval(): simple pole",1));
			}
		}
		if (((*_num2_p)*nx).is_integer()) {
			// half integer case
			if (nx.is_positive()) {
				// psi((2m+1)/2) -> 2/(2m+1) + 2/2m +...+ 2/1 - Euler - 2log(2)
				numeric rat = 0;
				for (numeric i = (nx+(*_num_1_p))*(*_num2_p); i>0; i-=(*_num2_p))
					rat += (*_num2_p)*i.inverse();
				return rat-Euler-_ex2*log(_ex2);
			} else {
				// use the recurrence relation
				//   psi(-m-1/2) == psi(-m-1/2+1) - 1 / (-m-1/2)
				// to relate psi(-m-1/2) to psi(1/2):
				//   psi(-m-1/2) == psi(1/2) + r
				// where r == ((-1/2)^(-1) + ... + (-m-1/2)^(-1))
				numeric recur = 0;
				for (numeric p = nx; p<0; ++p)
					recur -= pow(p, *_num_1_p);
				return recur+psi(_ex1_2);
			}
		}
		//  psi1_evalf should be called here once it becomes available
	}
	
	return this->hold();
}

ex psi1_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x)) {
		try {
			return psi(ex_to<numeric>(x));
		} catch (const dunno &e) { }
	}
	
	return this->hold();
}

ex psi1_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx psi(x) -> psi(1,x)
	return psi(_ex1, x);
}

ex psi1_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];
	// method:
	// Taylor series where there is no pole falls back to polygamma function
	// evaluation.
	// On a pole at -m use the recurrence relation
	//   psi(x) == psi(x+1) - 1/z
	// from which follows
	//   series(psi(x),x==-m,order) ==
	//   series(psi(x+m+1) - 1/x - 1/(x+1) - 1/(x+m)),x==-m,order);
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!arg_pt.info(info_flags::integer) || arg_pt.info(info_flags::positive))
		return basic::series(rel, order, options);
	// if we got here we have to care for a simple pole at -m:
	const numeric m = -ex_to<numeric>(arg_pt);
	ex recur;
	for (numeric p; p<=m; ++p)
		recur += power::power(arg+p,_ex_1);
	return (psi(arg+m+_ex1)-recur).series(rel, order, options);
}

void psi1_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\psi";
	inherited::do_print(c,level);
}

//////////
// Psi-functions (aka polygamma-functions)  psi(0,x)==psi(x)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(psi2_function,
		print_func<print_latex>(&psi2_function::do_print_latex))

/** Evaluation of polygamma-function psi(n,x). 
 *  Somebody ought to provide some good numerical evaluation some day... */
ex psi2_function::eval(int eval) const
{
	const ex& n = seq[0];
	const ex& x = seq[1];
	// psi(0,x) -> psi(x)
	if (n.is_zero())
		return psi(x);
	// psi(-1,x) -> log(tgamma(x))
	if (n.is_equal(_ex_1))
		return log(tgamma(x));
	if (n.info(info_flags::numeric) && n.info(info_flags::posint) &&
		x.info(info_flags::numeric)) {
		const numeric &nn = ex_to<numeric>(n);
		const numeric &nx = ex_to<numeric>(x);
		if (nx.is_integer()) {
			// integer case 
			if (nx.is_equal(*_num1_p))
				// use psi(n,1) == (-)^(n+1) * n! * zeta(n+1)
				return pow(*_num_1_p,nn+(*_num1_p))*factorial(nn)*zeta_function(nn+(*_num1_p));
			if (nx.is_positive()) {
				// use the recurrence relation
				//   psi(n,m) == psi(n,m+1) - (-)^n * n! / m^(n+1)
				// to relate psi(n,m) to psi(n,1):
				//   psi(n,m) == psi(n,1) + r
				// where r == (-)^n * n! * (1^(-n-1) + ... + (m-1)^(-n-1))
				numeric recur = 0;
				for (numeric p = 1; p<nx; ++p)
					recur += pow(p, -nn+(*_num_1_p));
				recur *= factorial(nn)*pow((*_num_1_p), nn);
				return recur+psi(n,_ex1);
			} else {
				// for non-positive integers there is a pole:
				throw (pole_error("psi2_eval(): pole",1));
			}
		}
		if (((*_num2_p)*nx).is_integer()) {
			// half integer case
			if (nx.is_equal(*_num1_2_p))
				// use psi(n,1/2) == (-)^(n+1) * n! * (2^(n+1)-1) * zeta(n+1)
				return pow(*_num_1_p,nn+(*_num1_p))*factorial(nn)*(pow(*_num2_p,nn+(*_num1_p)) + (*_num_1_p))*zeta_function(nn+(*_num1_p));
			if (nx.is_positive()) {
				const numeric m = nx - (*_num1_2_p);
				// use the multiplication formula
				//   psi(n,2*m) == (psi(n,m) + psi(n,m+1/2)) / 2^(n+1)
				// to revert to positive integer case
				return psi(n,(*_num2_p)*m)*pow((*_num2_p),nn+(*_num1_p))-psi(n,m);
			} else {
				// use the recurrence relation
				//   psi(n,-m-1/2) == psi(n,-m-1/2+1) - (-)^n * n! / (-m-1/2)^(n+1)
				// to relate psi(n,-m-1/2) to psi(n,1/2):
				//   psi(n,-m-1/2) == psi(n,1/2) + r
				// where r == (-)^(n+1) * n! * ((-1/2)^(-n-1) + ... + (-m-1/2)^(-n-1))
				numeric recur = 0;
				for (numeric p = nx; p<0; ++p)
					recur += pow(p, -nn+(*_num_1_p));
				recur *= factorial(nn)*pow(*_num_1_p, nn+(*_num_1_p));
				return recur+psi(n,_ex1_2);
			}
		}
		//  psi2_evalf should be called here once it becomes available
	}
	
	return this->hold();
}    

ex psi2_function::evalf(int eval) const
{
	const ex& n = seq[0];
	const ex& x = seq[1];
	if (is_exactly_a<numeric>(n) && is_exactly_a<numeric>(x)) {
		try {
			return psi(ex_to<numeric>(n),ex_to<numeric>(x));
		} catch (const dunno &e) { }
	}
	
	return this->hold();
}

ex psi2_function::pderivative(unsigned deriv_param) const
{
	const ex& n = seq[0];
	const ex& x = seq[1];
	GINAC_ASSERT(deriv_param<2);
	
	if (deriv_param==0) {
		// d/dn psi(n,x)
		throw(std::logic_error("cannot diff psi(n,x) with respect to n"));
	}
	// d/dx psi(n,x) -> psi(n+1,x)
	return psi(n+_ex1, x);
}

ex psi2_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& n = seq[0];
	const ex& arg = seq[1];
	// method:
	// Taylor series where there is no pole falls back to polygamma function
	// evaluation.
	// On a pole at -m use the recurrence relation
	//   psi(n,x) == psi(n,x+1) - (-)^n * n! / x^(n+1)
	// from which follows
	//   series(psi(x),x==-m,order) == 
	//   series(psi(x+m+1) - (-1)^n * n! * ((x)^(-n-1) + (x+1)^(-n-1) + ...
	//                                      ... + (x+m)^(-n-1))),x==-m,order);
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!arg_pt.info(info_flags::integer) || arg_pt.info(info_flags::positive))
		return basic::series(rel, order, options);
	// if we got here we have to care for a pole of order n+1 at -m:
	const numeric m = -ex_to<numeric>(arg_pt);
	ex recur;
	for (numeric p; p<=m; ++p)
		recur += power(arg+p,-n+_ex_1);
	recur *= factorial(n)*power(_ex_1,n);
	return (psi(n, arg+m+_ex1)-recur).series(rel, order, options);
}

void psi2_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\psi";
	inherited::do_print(c,level);
}

} // namespace GiNaC
