/** @file inifcns_exp.h
 *
 *  Interface to GiNaC's TODO */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __GINAC_INIFCNS_EXP_H__
#define __GINAC_INIFCNS_EXP_H__

#include "numeric.h"
#include "function.h"
#include "ex.h"

namespace GiNaC {

/** Eta function: log(a*b) == log(a) + log(b) + eta(a, b). */
//////////
// Eta function: eta(x,y) == log(x*y) - log(x) - log(y).
// This function is closely related to the unwinding number K, sometimes found
// in modern literature: K(z) == (z-log(exp(z)))/(2*Pi*I).
//////////
class eta_function : public function
{
	GINAC_DECLARE_FUNCTION_2P(eta_function)
public:
	virtual ex conjugate() const;
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1, typename T2> inline eta_function eta(const T1& x1, const T2& x2) { return eta_function(x1, x2); }

/** Exponential function. */
class exp_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(exp_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
protected:
	void do_print_csrc_float(const print_context& c, unsigned level) const;
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline exp_function exp(const T1& x1) { return exp_function(x1); }
inline exp_function exp(double x1) { return exp_function(x1); }
inline exp_function exp(float x1) { return exp_function(x1); }

/** Natural logarithm. */
class log_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(log_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_csrc_float(const print_context& c, unsigned level) const;
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline log_function log(const T1& x1) { return log_function(x1); }
inline log_function log(double x1) { return log_function(x1); }
inline log_function log(float x1) { return log_function(x1); }

/** Log-Gamma-function. */
class lgamma_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(lgamma_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline lgamma_function lgamma(const T1& x1) { return lgamma_function(x1); }

/** Gamma-function. */
class tgamma_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(tgamma_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline tgamma_function tgamma(const T1& x1) { return tgamma_function(x1); }

/** Beta-function. */
class beta_function : public function
{
	GINAC_DECLARE_FUNCTION_2P(beta_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1, typename T2> inline beta_function beta(const T1& x1, const T2& x2) { return beta_function(x1, x2); }

/** Psi-function (aka digamma-function). */
class psi1_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(psi1_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline psi1_function psi(const T1& x1) { return psi1_function(x1); }

/** Derivatives of Psi-function (aka polygamma-functions). */
class psi2_function : public function
{
	GINAC_DECLARE_FUNCTION_2P(psi2_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned deriv_param) const;
	virtual ex series(const relational& r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1, typename T2> inline psi2_function psi(const T1& x1, const T2& x2) { return psi2_function(x1, x2); }

} // namespace GiNaC

#endif // ndef __GINAC_INIFCNS_EXP_H__
