/** @file inifcns_trig.cpp
 *
 *  Implementation of trigonometric functions. */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "inifcns_trig.h"

#include "inifcns.h"
#include "inifcns_exp.h"
#include "ex.h"
#include "constant.h"
#include "numeric.h"
#include "power.h"
#include "operators.h"
#include "relational.h"
#include "symbol.h"
#include "pseries.h"
#include "utils.h"

#include <stdexcept>
#include <string>

namespace GiNaC {

////////////////////////////////////////////////////////////////////////////////
// sine (trigonometric function)
////////////////////////////////////////////////////////////////////////////////

GINAC_IMPLEMENT_FUNCTION_OPT(sin_function,
		print_func<print_latex>(&sin_function::do_print_latex))

ex sin_function::eval(int level) const
{
	const ex& x = seq[0];
	// sin(n/d*Pi) -> { all known non-nested radicals }
	const ex SixtyExOverPi = _ex60*x/Pi;
	ex sign = _ex1;
	if (SixtyExOverPi.info(info_flags::integer)) {
		numeric z = mod(ex_to<numeric>(SixtyExOverPi),*_num120_p);
		if (z>=*_num60_p) {
			// wrap to interval [0, Pi)
			z -= *_num60_p;
			sign = _ex_1;
		}
		if (z>*_num30_p) {
			// wrap to interval [0, Pi/2)
			z = *_num60_p-z;
		}
		if (z.is_equal(*_num0_p))  // sin(0)       -> 0
			return _ex0;
		if (z.is_equal(*_num5_p))  // sin(Pi/12)   -> sqrt(6)/4*(1-sqrt(3)/3)
			return sign*_ex1_4*sqrt(_ex6)*(_ex1+_ex_1_3*sqrt(_ex3));
		if (z.is_equal(*_num6_p))  // sin(Pi/10)   -> sqrt(5)/4-1/4
			return sign*(_ex1_4*sqrt(_ex5)+_ex_1_4);
		if (z.is_equal(*_num10_p)) // sin(Pi/6)    -> 1/2
			return sign*_ex1_2;
		if (z.is_equal(*_num15_p)) // sin(Pi/4)    -> sqrt(2)/2
			return sign*_ex1_2*sqrt(_ex2);
		if (z.is_equal(*_num18_p)) // sin(3/10*Pi) -> sqrt(5)/4+1/4
			return sign*(_ex1_4*sqrt(_ex5)+_ex1_4);
		if (z.is_equal(*_num20_p)) // sin(Pi/3)    -> sqrt(3)/2
			return sign*_ex1_2*sqrt(_ex3);
		if (z.is_equal(*_num25_p)) // sin(5/12*Pi) -> sqrt(6)/4*(1+sqrt(3)/3)
			return sign*_ex1_4*sqrt(_ex6)*(_ex1+_ex1_3*sqrt(_ex3));
		if (z.is_equal(*_num30_p)) // sin(Pi/2)    -> 1
			return sign;
	}

	if (is_exactly_a<asin_function>(x)) {
		// sin(asin(x)) -> x
		return x.op(0);
	} else if (is_exactly_a<acos_function>(x)) {
		// sin(acos(x)) -> sqrt(1-x^2)
		return sqrt(_ex1-power::power(x.op(0),_ex2));
	} else if (is_exactly_a<atan_function>(x)) {
		const ex& t = x.op(0);
		// sin(atan(x)) -> x/sqrt(1+x^2)
		return t*power::power(_ex1+power::power(t,_ex2),_ex_1_2);
	}
	
	// sin(float) -> float
	if (x.info(info_flags::numeric) && !x.info(info_flags::crational))
		return sin(ex_to<numeric>(x));

	// sin() is odd
	if (x.info(info_flags::negative))
		return -sin(-x);
	
	return this->hold();
}

ex sin_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return sin(ex_to<numeric>(x));
	
	return this->hold();
}

ex sin_function::pderivative(unsigned deriv_param) const
{
    GINAC_ASSERT(deriv_param==0);
	// d/dx sin(x) -> cos(x)
	return cos(seq[0]);
}

void sin_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\sin";
	inherited::do_print(c,level);
}

//////////
// cosine (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(cos_function,
		print_func<print_latex>(&cos_function::do_print_latex))

ex cos_function::eval(int level) const
{
	const ex& x = seq[0];
	// cos(n/d*Pi) -> { all known non-nested radicals }
	const ex SixtyExOverPi = _ex60*x/Pi;
	ex sign = _ex1;
	if (SixtyExOverPi.info(info_flags::integer)) {
		numeric z = mod(ex_to<numeric>(SixtyExOverPi),*_num120_p);
		if (z>=*_num60_p) {
			// wrap to interval [0, Pi)
			z = *_num120_p-z;
		}
		if (z>=*_num30_p) {
			// wrap to interval [0, Pi/2)
			z = *_num60_p-z;
			sign = _ex_1;
		}
		if (z.is_equal(*_num0_p))  // cos(0)       -> 1
			return sign;
		if (z.is_equal(*_num5_p))  // cos(Pi/12)   -> sqrt(6)/4*(1+sqrt(3)/3)
			return sign*_ex1_4*sqrt(_ex6)*(_ex1+_ex1_3*sqrt(_ex3));
		if (z.is_equal(*_num10_p)) // cos(Pi/6)    -> sqrt(3)/2
			return sign*_ex1_2*sqrt(_ex3);
		if (z.is_equal(*_num12_p)) // cos(Pi/5)    -> sqrt(5)/4+1/4
			return sign*(_ex1_4*sqrt(_ex5)+_ex1_4);
		if (z.is_equal(*_num15_p)) // cos(Pi/4)    -> sqrt(2)/2
			return sign*_ex1_2*sqrt(_ex2);
		if (z.is_equal(*_num20_p)) // cos(Pi/3)    -> 1/2
			return sign*_ex1_2;
		if (z.is_equal(*_num24_p)) // cos(2/5*Pi)  -> sqrt(5)/4-1/4x
			return sign*(_ex1_4*sqrt(_ex5)+_ex_1_4);
		if (z.is_equal(*_num25_p)) // cos(5/12*Pi) -> sqrt(6)/4*(1-sqrt(3)/3)
			return sign*_ex1_4*sqrt(_ex6)*(_ex1+_ex_1_3*sqrt(_ex3));
		if (z.is_equal(*_num30_p)) // cos(Pi/2)    -> 0
			return _ex0;
	}

	if (is_exactly_a<acos_function>(x)) {
		// cos(acos(x)) -> x
		return x.op(0);
	} else if (is_exactly_a<asin_function>(x)) {
		// cos(asin(x)) -> sqrt(1-x^2)
		return sqrt(_ex1-power::power(x.op(0),_ex2));
	} else if (is_exactly_a<atan_function>(x)) {
		// cos(atan(x)) -> 1/sqrt(1+x^2)
		return power::power(_ex1+power::power(x.op(0),_ex2),_ex_1_2);
	}
	
	// cos(float) -> float
	if (x.info(info_flags::numeric) && !x.info(info_flags::crational))
		return cos(ex_to<numeric>(x));
	
	// cos() is even
	if (x.info(info_flags::negative))
		return cos(-x);
	
	return this->hold();
}

ex cos_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return cos(ex_to<numeric>(x));
	
	return this->hold();
}

ex cos_function::pderivative(unsigned deriv_param) const
{
	GINAC_ASSERT(deriv_param==0);

	// d/dx cos(x) -> -sin(x)
	return -sin(seq[0]);
}

void cos_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\cos";
	inherited::do_print(c,level);
}

//////////
// tangent (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(tan_function,
		print_func<print_latex>(&tan_function::do_print_latex))

ex tan_function::eval(int level) const
{
	const ex& x = seq[0];
	// tan(n/d*Pi) -> { all known non-nested radicals }
	const ex SixtyExOverPi = _ex60*x/Pi;
	ex sign = _ex1;
	if (SixtyExOverPi.info(info_flags::integer)) {
		numeric z = mod(ex_to<numeric>(SixtyExOverPi),*_num60_p);
		if (z>=*_num60_p) {
			// wrap to interval [0, Pi)
			z -= *_num60_p;
		}
		if (z>=*_num30_p) {
			// wrap to interval [0, Pi/2)
			z = *_num60_p-z;
			sign = _ex_1;
		}
		if (z.is_equal(*_num0_p))  // tan(0)       -> 0
			return _ex0;
		if (z.is_equal(*_num5_p))  // tan(Pi/12)   -> 2-sqrt(3)
			return sign*(_ex2-sqrt(_ex3));
		if (z.is_equal(*_num10_p)) // tan(Pi/6)    -> sqrt(3)/3
			return sign*_ex1_3*sqrt(_ex3);
		if (z.is_equal(*_num15_p)) // tan(Pi/4)    -> 1
			return sign;
		if (z.is_equal(*_num20_p)) // tan(Pi/3)    -> sqrt(3)
			return sign*sqrt(_ex3);
		if (z.is_equal(*_num25_p)) // tan(5/12*Pi) -> 2+sqrt(3)
			return sign*(sqrt(_ex3)+_ex2);
		if (z.is_equal(*_num30_p)) // tan(Pi/2)    -> infinity
			throw (pole_error("tan_eval(): simple pole",1));
	}

	if (is_exactly_a<atan_function>(x)) {
		// tan(atan(x)) -> x
		return x.op(0);
	} else if (is_exactly_a<asin_function>(x)) {
		// tan(asin(x)) -> x/sqrt(1+x^2)
		const ex& t = x.op(0);
		return t*power::power(_ex1-power::power(t,_ex2),_ex_1_2);
	} else if (is_exactly_a<acos_function>(x)) {
		// tan(acos(x)) -> sqrt(1-x^2)/x
		const ex& t = x.op(0);
		return power::power(t,_ex_1)*sqrt(_ex1-power::power(t,_ex2));
	}
	
	// tan(float) -> float
	if (x.info(info_flags::numeric) && !x.info(info_flags::crational)) {
		return tan(ex_to<numeric>(x));
	}
	
	// tan() is odd
	if (x.info(info_flags::negative)) {
		return -tan(-x);
	}
	
	return this->hold();
}

ex tan_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return tan(ex_to<numeric>(x));
	
	return this->hold();
}

ex tan_function::pderivative(unsigned deriv_param) const
{
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx tan(x) -> 1+tan(x)^2;
	return (_ex1+power::power(tan(seq[0]),_ex2));
}

ex tan_function::series(const relational& rel, int order, unsigned options) const
{
	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	// method:
	// Taylor series where there is no pole falls back to tan_deriv.
	// On a pole simply expand sin(x)/cos(x).
	const ex x_pt = seq[0].subs(rel, subs_options::no_pattern);
	if (!(2*x_pt/Pi).info(info_flags::odd)) {
		return basic::series(rel, order, options);
	}
	// if we got here we have to care for a simple pole
	const ex& x = seq[0];
	return (sin(x)/cos(x)).series(rel, order, options);
}

void tan_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\tan";
	inherited::do_print(c,level);
}

//////////
// inverse sine (arc sine)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(asin_function,
		print_func<print_latex>(&asin_function::do_print_latex))

ex asin_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// asin(0) -> 0
		if (x.is_zero())
			return x;

		// asin(1/2) -> Pi/6
		if (x.is_equal(_ex1_2))
			return numeric(1,6)*Pi;

		// asin(1) -> Pi/2
		if (x.is_equal(_ex1))
			return _ex1_2*Pi;

		// asin(-1/2) -> -Pi/6
		if (x.is_equal(_ex_1_2))
			return numeric(-1,6)*Pi;

		// asin(-1) -> -Pi/2
		if (x.is_equal(_ex_1))
			return _ex_1_2*Pi;

		// asin(float) -> float
		if (!x.info(info_flags::crational))
			return asin(ex_to<numeric>(x));

		// asin() is odd
		if (x.info(info_flags::negative))
			return -asin(-x);
	}
	
	return this->hold();
}

ex asin_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return asin(ex_to<numeric>(x));
	
	return this->hold();
}

ex asin_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx asin(x) -> 1/sqrt(1-x^2)
	return power::power(1-power::power(x,_ex2),_ex_1_2);
}

void asin_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arcsin";
	inherited::do_print(c,level);
}

//////////
// inverse cosine (arc cosine)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(acos_function,
		print_func<print_latex>(&acos_function::do_print_latex))

ex acos_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// acos(1) -> 0
		if (x.is_equal(_ex1))
			return _ex0;

		// acos(1/2) -> Pi/3
		if (x.is_equal(_ex1_2))
			return _ex1_3*Pi;

		// acos(0) -> Pi/2
		if (x.is_zero())
			return _ex1_2*Pi;

		// acos(-1/2) -> 2/3*Pi
		if (x.is_equal(_ex_1_2))
			return numeric(2,3)*Pi;

		// acos(-1) -> Pi
		if (x.is_equal(_ex_1))
			return Pi;

		// acos(float) -> float
		if (!x.info(info_flags::crational))
			return acos(ex_to<numeric>(x));

		// acos(-x) -> Pi-acos(x)
		if (x.info(info_flags::negative))
			return Pi-acos(-x);
	}
	
	return this->hold();
}

ex acos_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return acos(ex_to<numeric>(x));
	
	return this->hold();
}

ex acos_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx acos(x) -> -1/sqrt(1-x^2)
	return -power::power(1-power::power(x,_ex2),_ex_1_2);
}

void acos_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arccos";
	inherited::do_print(c,level);
}

//////////
// inverse tangent (arc tangent)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(atan_function,
		print_func<print_latex>(&atan_function::do_print_latex))

ex atan_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// atan(0) -> 0
		if (x.is_zero())
			return _ex0;

		// atan(1) -> Pi/4
		if (x.is_equal(_ex1))
			return _ex1_4*Pi;

		// atan(-1) -> -Pi/4
		if (x.is_equal(_ex_1))
			return _ex_1_4*Pi;

		if (x.is_equal(I) || x.is_equal(-I))
			throw (pole_error("atan_eval(): logarithmic pole",0));

		// atan(float) -> float
		if (!x.info(info_flags::crational))
			return atan(ex_to<numeric>(x));

		// atan() is odd
		if (x.info(info_flags::negative))
			return -atan(-x);
	}
	
	return this->hold();
}

ex atan_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return atan(ex_to<numeric>(x));
	
	return this->hold();
}

ex atan_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);

	// d/dx atan(x) -> 1/(1+x^2)
	return power::power(_ex1+power::power(x,_ex2), _ex_1);
}

ex atan_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];
	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	// method:
	// Taylor series where there is no pole or cut falls back to atan_deriv.
	// There are two branch cuts, one runnig from I up the imaginary axis and
	// one running from -I down the imaginary axis.  The points I and -I are
	// poles.
	// On the branch cuts and the poles series expand
	//     (log(1+I*x)-log(1-I*x))/(2*I)
	// instead.
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!(I*arg_pt).info(info_flags::real))
		return basic::series(rel, order, options);
	if ((I*arg_pt).info(info_flags::real) && abs(I*arg_pt)<_ex1)
		return basic::series(rel, order, options);
	// care for the poles, using the defining formula for atan()...
	if (arg_pt.is_equal(I) || arg_pt.is_equal(-I))
		return ((log(1+I*arg)-log(1-I*arg))/(2*I)).series(rel, order, options);
	if (!(options & series_options::suppress_branchcut)) {
		// method:
		// This is the branch cut: assemble the primitive series manually and
		// then add the corresponding complex step function.
		const symbol& s = ex_to<symbol>(rel.lhs());
		const ex& point = rel.rhs();
		const symbol foo;
		const ex replarg = atan(arg).series(s==foo, order).subs(foo==point, subs_options::no_pattern);
		ex Order0correction = replarg.op(0)+csgn(arg)*Pi*_ex_1_2;
		if ((I*arg_pt)<_ex0)
			Order0correction += log((I*arg_pt+_ex_1)/(I*arg_pt+_ex1))*I*_ex_1_2;
		else
			Order0correction += log((I*arg_pt+_ex1)/(I*arg_pt+_ex_1))*I*_ex1_2;
		epvector seq;
		seq.push_back(expair(Order0correction, _ex0));
		seq.push_back(expair(Order(_ex1), order));
		return (replarg - pseries(rel, seq)).series(rel, order);
	}
	return basic::series(rel, order, options);
}

void atan_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arctan";
	inherited::do_print(c,level);
}

//////////
// inverse tangent (atan2(y,x))
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(atan2_function,
		print_func<print_latex>(&atan2_function::do_print_latex))

ex atan2_function::eval(int level) const
{
	const ex& y = seq[0];
	const ex& x = seq[1];

	if (y.info(info_flags::numeric) && x.info(info_flags::numeric)) {

		if (y.is_zero()) {

			// atan(0, 0) -> 0
			if (x.is_zero())
				return _ex0;

			// atan(0, x), x real and positive -> 0
			if (x.info(info_flags::positive))
				return _ex0;

			// atan(0, x), x real and negative -> -Pi
			if (x.info(info_flags::negative))
				return _ex_1*Pi;
		}

		if (x.is_zero()) {

			// atan(y, 0), y real and positive -> Pi/2
			if (y.info(info_flags::positive))
				return _ex1_2*Pi;

			// atan(y, 0), y real and negative -> -Pi/2
			if (y.info(info_flags::negative))
				return _ex_1_2*Pi;
		}

		if (y.is_equal(x)) {

			// atan(y, y), y real and positive -> Pi/4
			if (y.info(info_flags::positive))
				return _ex1_4*Pi;

			// atan(y, y), y real and negative -> -3/4*Pi
			if (y.info(info_flags::negative))
				return numeric(-3, 4)*Pi;
		}

		if (y.is_equal(-x)) {

			// atan(y, -y), y real and positive -> 3*Pi/4
			if (y.info(info_flags::positive))
				return numeric(3, 4)*Pi;

			// atan(y, -y), y real and negative -> -Pi/4
			if (y.info(info_flags::negative))
				return _ex_1_4*Pi;
		}

		// atan(float, float) -> float
		if (!y.info(info_flags::crational) && !x.info(info_flags::crational))
			return atan(ex_to<numeric>(y), ex_to<numeric>(x));

		// atan(real, real) -> atan(y/x) +/- Pi
		if (y.info(info_flags::real) && x.info(info_flags::real)) {
			if (x.info(info_flags::positive))
				return atan(y/x);
			else if(y.info(info_flags::positive))
				return atan(y/x)+Pi;
			else
				return atan(y/x)-Pi;
		}
	}

	return this->hold();
}    

ex atan2_function::evalf(int level) const
{
	const ex& y = seq[0];
	const ex& x = seq[1];
	if (is_exactly_a<numeric>(y) && is_exactly_a<numeric>(x))
		return atan(ex_to<numeric>(y), ex_to<numeric>(x));
	
	return this->hold();
}

ex atan2_function::pderivative(unsigned deriv_param) const
{
	const ex& y = seq[0];
	const ex& x = seq[1];
	GINAC_ASSERT(deriv_param<2);
	
	if (deriv_param==0) {
		// d/dy atan(y,x)
		return x*power::power(power::power(x,_ex2)+power::power(y,_ex2),_ex_1);
	}
	// d/dx atan(y,x)
	return -y*power::power(power::power(x,_ex2)+power::power(y,_ex2),_ex_1);
}

void atan2_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arctan_2";
	inherited::do_print(c,level);
}

//////////
// hyperbolic sine (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(sinh_function,
		print_func<print_latex>(&sinh_function::do_print_latex))

ex sinh_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// sinh(0) -> 0
		if (x.is_zero())
			return _ex0;        

		// sinh(float) -> float
		if (!x.info(info_flags::crational))
			return sinh(ex_to<numeric>(x));

		// sinh() is odd
		if (x.info(info_flags::negative))
			return -sinh(-x);
	}
	
	if ((x/Pi).info(info_flags::numeric) &&
		ex_to<numeric>(x/Pi).real().is_zero())  // sinh(I*x) -> I*sin(x)
		return I*sin(x/I);
	
	const ex& t = x.op(0);
	if (is_exactly_a<asinh_function>(x)) {
		// sinh(asinh(x)) -> x
		return t;
	} else if (is_exactly_a<acosh_function>(x)) {
		// sinh(acosh(x)) -> sqrt(x-1) * sqrt(x+1)
		return sqrt(t-_ex1)*sqrt(t+_ex1);
	} else if (is_exactly_a<atanh_function>(x)) {
		// sinh(atanh(x)) -> x/sqrt(1-x^2)
		return t*power::power(_ex1-power::power(t,_ex2),_ex_1_2);
	}
	
	return this->hold();
}

ex sinh_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return sinh(ex_to<numeric>(x));
	
	return this->hold();
}

ex sinh_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx sinh(x) -> cosh(x)
	return cosh(x);
}

void sinh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arcsin";
	inherited::do_print(c,level);
}

//////////
// hyperbolic cosine (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(cosh_function,
		print_func<print_latex>(&cosh_function::do_print_latex))

ex cosh_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// cosh(0) -> 1
		if (x.is_zero())
			return _ex1;

		// cosh(float) -> float
		if (!x.info(info_flags::crational))
			return cosh(ex_to<numeric>(x));

		// cosh() is even
		if (x.info(info_flags::negative))
			return cosh(-x);
	}
	
	if ((x/Pi).info(info_flags::numeric) &&
		ex_to<numeric>(x/Pi).real().is_zero())  // cosh(I*x) -> cos(x)
		return cos(x/I);
	
	if (is_exactly_a<acosh_function>(x)) {
		// cosh(acosh(x)) -> x
		return x.op(0);
	} else if (is_exactly_a<asinh_function>(x)) {
		// cosh(asinh(x)) -> sqrt(1+x^2)
		return sqrt(_ex1+power::power(x.op(0),_ex2));
	} else if (is_exactly_a<atanh_function>(x)) {
		// cosh(atanh(x)) -> 1/sqrt(1-x^2)
		return power::power(_ex1-power::power(x.op(0),_ex2),_ex_1_2);
	}
	
	return this->hold();
}

ex cosh_function::evalf(int level) const
{
	const ex& x = seq[0];

	if (is_exactly_a<numeric>(x))
		return cosh(ex_to<numeric>(x));
	
	return this->hold();
}

ex cosh_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx cosh(x) -> sinh(x)
	return sinh(x);
}

void cosh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arccos";
	inherited::do_print(c,level);
}

//////////
// hyperbolic tangent (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(tanh_function,
		print_func<print_latex>(&tanh_function::do_print_latex))

ex tanh_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// tanh(0) -> 0
		if (x.is_zero())
			return _ex0;

		// tanh(float) -> float
		if (!x.info(info_flags::crational))
			return tanh(ex_to<numeric>(x));

		// tanh() is odd
		if (x.info(info_flags::negative))
			return -tanh(-x);
	}
	
	if ((x/Pi).info(info_flags::numeric) &&
		ex_to<numeric>(x/Pi).real().is_zero())  // tanh(I*x) -> I*tan(x);
		return I*tan(x/I);
	
	if (is_exactly_a<atanh_function>(x)) {
		// tanh(atanh(x)) -> x
		return x.op(0);
	} else if (is_exactly_a<asinh_function>(x)) {
		// tanh(asinh(x)) -> x/sqrt(1+x^2)
		const ex& t = x.op(0);
		return t*power::power(_ex1+power::power(t,_ex2),_ex_1_2);
	} else if (is_exactly_a<acosh_function>(x)) {
		// tanh(acosh(x)) -> sqrt(x-1)*sqrt(x+1)/x
		const ex& t = x.op(0);
		return sqrt(t-_ex1)*sqrt(t+_ex1)*power::power(t,_ex_1);
	}
	
	return this->hold();
}

ex tanh_function::evalf(int level) const
{
	const ex& x = seq[0];
	if (is_exactly_a<numeric>(x))
		return tanh(ex_to<numeric>(x));
	
	return this->hold();
}

ex tanh_function::pderivative(unsigned deriv_param) const
{
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx tanh(x) -> 1-tanh(x)^2
	return _ex1-power::power(tanh(seq[0]),_ex2);
}

ex tanh_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& x = seq[0];

	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	// method:
	// Taylor series where there is no pole falls back to tanh_deriv.
	// On a pole simply expand sinh(x)/cosh(x).
	const ex x_pt = x.subs(rel, subs_options::no_pattern);
	if (!(2*I*x_pt/Pi).info(info_flags::odd))
		return basic::series(rel, order, options);
	// if we got here we have to care for a simple pole
	return (sinh(x)/cosh(x)).series(rel, order, options);
}

void tanh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arctan";
	inherited::do_print(c,level);
}

//////////
// inverse hyperbolic sine (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(asinh_function,
		print_func<print_latex>(&asinh_function::do_print_latex))

ex asinh_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// asinh(0) -> 0
		if (x.is_zero())
			return _ex0;

		// asinh(float) -> float
		if (!x.info(info_flags::crational))
			return asinh(ex_to<numeric>(x));

		// asinh() is odd
		if (x.info(info_flags::negative))
			return -asinh(-x);
	}
	
	return this->hold();
}

ex asinh_function::evalf(int level) const
{
	const ex& x = seq[0];

	if (is_exactly_a<numeric>(x))
		return asinh(ex_to<numeric>(x));
	
	return this->hold();
}

ex asinh_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];

	GINAC_ASSERT(deriv_param==0);
	
	// d/dx asinh(x) -> 1/sqrt(1+x^2)
	return power::power(_ex1+power::power(x,_ex2),_ex_1_2);
}

void asinh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arcsinh";
	inherited::do_print(c,level);
}

//////////
// inverse hyperbolic cosine (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(acosh_function,
		print_func<print_latex>(&acosh_function::do_print_latex))

ex acosh_function::eval(int level) const
{
	const ex& x = seq[0];

	if (x.info(info_flags::numeric)) {

		// acosh(0) -> Pi*I/2
		if (x.is_zero())
			return Pi*I*numeric(1,2);

		// acosh(1) -> 0
		if (x.is_equal(_ex1))
			return _ex0;

		// acosh(-1) -> Pi*I
		if (x.is_equal(_ex_1))
			return Pi*I;

		// acosh(float) -> float
		if (!x.info(info_flags::crational))
			return acosh(ex_to<numeric>(x));

		// acosh(-x) -> Pi*I-acosh(x)
		if (x.info(info_flags::negative))
			return Pi*I-acosh(-x);
	}
	
	return this->hold();
}

ex acosh_function::evalf(int level) const
{
	const ex& x = seq[0];

	if (is_exactly_a<numeric>(x))
		return acosh(ex_to<numeric>(x));
	
	return this->hold();
}

ex acosh_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];

	GINAC_ASSERT(deriv_param==0);
	
	// d/dx acosh(x) -> 1/(sqrt(x-1)*sqrt(x+1))
	return power::power(x+_ex_1,_ex_1_2)*power::power(x+_ex1,_ex_1_2);
}

void acosh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arccosh";
	inherited::do_print(c,level);
}

//////////
// inverse hyperbolic tangent (trigonometric function)
//////////

GINAC_IMPLEMENT_FUNCTION_OPT(atanh_function,
		print_func<print_latex>(&atanh_function::do_print_latex))

ex atanh_function::eval(int level) const
{
	const ex& x = seq[0];
	if (x.info(info_flags::numeric)) {

		// atanh(0) -> 0
		if (x.is_zero())
			return _ex0;

		// atanh({+|-}1) -> throw
		if (x.is_equal(_ex1) || x.is_equal(_ex_1))
			throw (pole_error("atanh_eval(): logarithmic pole",0));

		// atanh(float) -> float
		if (!x.info(info_flags::crational))
			return atanh(ex_to<numeric>(x));

		// atanh() is odd
		if (x.info(info_flags::negative))
			return -atanh(-x);
	}
	
	return this->hold();
}

ex atanh_function::evalf(int level) const
{
	const ex& x = seq[0];

	if (is_exactly_a<numeric>(x))
		return atanh(ex_to<numeric>(x));
	
	return this->hold();
}

ex atanh_function::pderivative(unsigned deriv_param) const
{
	const ex& x = seq[0];
	GINAC_ASSERT(deriv_param==0);
	
	// d/dx atanh(x) -> 1/(1-x^2)
	return power::power(_ex1-power::power(x,_ex2),_ex_1);
}

ex atanh_function::series(const relational& rel, int order, unsigned options) const
{
	const ex& arg = seq[0];

	GINAC_ASSERT(is_a<symbol>(rel.lhs()));
	// method:
	// Taylor series where there is no pole or cut falls back to atanh_deriv.
	// There are two branch cuts, one runnig from 1 up the real axis and one
	// one running from -1 down the real axis.  The points 1 and -1 are poles
	// On the branch cuts and the poles series expand
	//     (log(1+x)-log(1-x))/2
	// instead.
	const ex arg_pt = arg.subs(rel, subs_options::no_pattern);
	if (!(arg_pt).info(info_flags::real))
		return basic::series(rel, order, options);
	if ((arg_pt).info(info_flags::real) && abs(arg_pt)<_ex1)
		return basic::series(rel, order, options);
	// care for the poles, using the defining formula for atanh()...
	if (arg_pt.is_equal(_ex1) || arg_pt.is_equal(_ex_1))
		return ((log(_ex1+arg)-log(_ex1-arg))*_ex1_2).series(rel, order, options);
	// ...and the branch cuts (the discontinuity at the cut being just I*Pi)
	if (!(options & series_options::suppress_branchcut)) {
 		// method:
 		// This is the branch cut: assemble the primitive series manually and
 		// then add the corresponding complex step function.
 		const symbol &s = ex_to<symbol>(rel.lhs());
 		const ex &point = rel.rhs();
 		const symbol foo;
 		const ex replarg = atanh(arg).series(s==foo, order).subs(foo==point, subs_options::no_pattern);
		ex Order0correction = replarg.op(0)+csgn(I*arg)*Pi*I*_ex1_2;
		if (arg_pt<_ex0)
			Order0correction += log((arg_pt+_ex_1)/(arg_pt+_ex1))*_ex1_2;
		else
			Order0correction += log((arg_pt+_ex1)/(arg_pt+_ex_1))*_ex_1_2;
 		epvector seq;
		seq.push_back(expair(Order0correction, _ex0));
 		seq.push_back(expair(Order(_ex1), order));
 		return (replarg - pseries(rel, seq)).series(rel, order);
	}
	return basic::series(rel, order, options);
}

void atanh_function::do_print_latex(const print_context& c, unsigned level) const
{
	c.s << "\\arctanh";
	inherited::do_print(c,level);
}

} // namespace GiNaC
