/** @file inifcns_trig.h
 *
 *  Interface to GiNaC's initially known trigonometric functions. */

/*
 *  GiNaC Copyright (C) 1999-2006 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __GINAC_INIFCNS_TRIG_H__
#define __GINAC_INIFCNS_TRIG_H__

#include "ex.h"
#include "function.h"

namespace GiNaC {

/** Sine. */

class sin_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(sin_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline sin_function sin(const T1& x1) { return sin_function(x1); }
inline sin_function sin(double x1);
inline sin_function sin(float x1);

/** Cosine. */

class cos_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(cos_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline cos_function cos(const T1& x1) { return cos_function(x1); }
inline cos_function cos(double x1);
inline cos_function cos(float x1);

/** Tangent. */

class tan_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(tan_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
	virtual ex series(const relational & r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline tan_function tan(const T1& x1) { return tan_function(x1); }
inline tan_function tan(double x1);
inline tan_function tan(float x1);

/** Inverse sine (arc sine). */

class asin_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(asin_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline asin_function asin(const T1& x1) { return asin_function(x1); }
inline asin_function asin(double x1);
inline asin_function asin(float x1);

/** Inverse cosine (arc cosine). */

class acos_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(acos_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline acos_function acos(const T1& x1) { return acos_function(x1); }
inline acos_function acos(double x1);
inline acos_function acos(float x1);

/** Inverse tangent (arc tangent). */

class atan_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(atan_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
	virtual ex series(const relational & r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline atan_function atan(const T1& x1) { return atan_function(x1); }
inline atan_function atan(double x1);
inline atan_function atan(float x1);

/** Inverse tangent with two arguments. */

class atan2_function : public function
{
	GINAC_DECLARE_FUNCTION_2P(atan2_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1, typename T2> inline atan2_function atan2(const T1& x1, const T2& x2) { return atan2_function(x1, x2); }
inline atan2_function atan2(double x1, double x2);
inline atan2_function atan2(float x1, float x2);

/** Hyperbolic Sine. */

class sinh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(sinh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline sinh_function sinh(const T1& x1) { return sinh_function(x1); }
inline sinh_function sinh(double x1);
inline sinh_function sinh(float x1);

/** Hyperbolic Cosine. */

class cosh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(cosh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline cosh_function cosh(const T1& x1) { return cosh_function(x1); }
inline cosh_function cosh(double x1);
inline cosh_function cosh(float x1);

/** Hyperbolic Tangent. */

class tanh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(tanh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
	virtual ex series(const relational & r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline tanh_function tanh(const T1& x1) { return tanh_function(x1); }
inline tanh_function tanh(double x1);
inline tanh_function tanh(float x1);

/** Inverse hyperbolic Sine (area hyperbolic sine). */

class asinh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(asinh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline asinh_function asinh(const T1& x1) { return asinh_function(x1); }
inline asinh_function asinh(double x1);
inline asinh_function asinh(float x1);

/** Inverse hyperbolic Cosine (area hyperbolic cosine). */

class acosh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(acosh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline acosh_function acosh(const T1& x1) { return acosh_function(x1); }
inline acosh_function acosh(double x1);
inline acosh_function acosh(float x1);

/** Inverse hyperbolic Tangent (area hyperbolic tangent). */

class atanh_function : public function
{
	GINAC_DECLARE_FUNCTION_1P(atanh_function)
public:
	virtual ex eval(int level = 0) const;
	virtual ex evalf(int level = 0) const;
	virtual ex pderivative(unsigned diff_param) const;
	virtual ex series(const relational & r, int order, unsigned options = 0) const;
protected:
	void do_print_latex(const print_context& c, unsigned level) const;
};

template<typename T1> inline atanh_function atanh(const T1& x1) { return atanh_function(x1); }
inline atanh_function atanh(double x1);
inline atanh_function atanh(float x1);

} // namespace GiNaC

#endif // ifndef __GINAC_INIFCNS_TRIG_H__
